from etcd import EtcdKeyNotFound

from distconfig.backends.base import BaseBackend
from distconfig.backends.execution_context import ThreadingExecutionContext


class EtcdBackend(BaseBackend):
    """Etcd backend implementation.

    :param client: Instance of :class:`etcd.Client`.
    :param execution_context: Instance of :class:`distconfig.backends.execution_context.ExecutionContext`
    """

    def __init__(self, client, execution_context=ThreadingExecutionContext(), **kwargs):
        super(EtcdBackend, self).__init__(**kwargs)
        self._client = client
        self._execution_context = execution_context
        self._watching = set()

    def get_raw(self, key):
        result = self._get_backend_data(key)
        self._add_watcher(key)
        return result

    def _get_backend_data(self, key):
        try:
            result = self._client.get(key)
        except EtcdKeyNotFound:
            return
        else:
            return result.value

    def _add_watcher(self, key):
        if key not in self._watching:
            self._watching.add(key)
            self._execution_context.run(self._watch_for_changes, key)

    def _watch_for_changes(self, key):
        index = None
        while 1:
            try:
                response = self._client.watch(key, index=index)
            except Exception as ex:
                self._logger.error('exception raised while listening on etcd changes (re-launching watcher): %s', ex)
            else:
                if index is None:
                    index = response.etcd_index
                else:
                    index += 1
                self._notify_listeners(response.value)



import abc
import threading


class ExecutionContext(metaclass=abc.ABCMeta):
    """Base abstract execution context class."""

    @abc.abstractmethod
    def run(self, func, *args, **kwargs):
        pass


class GeventExecutionContext(ExecutionContext):
    """Execution context that run background function as a Greenlet.

    gevent monkey patching must be done by user.
    """

    def run(self, func, *args, **kwargs):
        """Run given function in a Greenlet."""
        import gevent

        gevent.spawn(func, *args, **kwargs)
        gevent.sleep()


class ThreadingExecutionContext(ExecutionContext):
    """Execution context that run background function as a OS Thread."""

    def run(self, func, *args, **kwargs):
        """Run given function in a daemon OS thread."""
        thread = threading.Thread(target=func, args=args, kwargs=kwargs)
        thread.daemon = True
        thread.start()



import http.client
import os
import socket
import unittest

import etcd

from distconfig.tests.integration.base import _BackendTestCase
from distconfig.api import Proxy


EPHEMERAL_TEST_DATA_TTL = 60
ETCD_ENDPOINT_IP = os.environ.get('ETCD_ENDPOINT_IP', '127.0.0.1')
ETCD_ENDPOINT_PORT = int(os.environ.get('ETCD_ENDPOINT_PORT', 4001))


def etcd_running():
    conn = http.client.HTTPConnection(ETCD_ENDPOINT_IP, port=ETCD_ENDPOINT_PORT)
    try:
        conn.request('HEAD', '/')
    except (socket.error, socket.gaierror):
        return False
    else:
        return True


@unittest.skipUnless(etcd_running(), "Can't connect to Etcd on %s:%s" % (ETCD_ENDPOINT_IP, ETCD_ENDPOINT_PORT))
class EtcdBackendTest(_BackendTestCase):

    path = 'distconfig/testing/config'
    client = etcd.Client(ETCD_ENDPOINT_IP, ETCD_ENDPOINT_PORT)
    proxy = Proxy.configure(
        'distconfig.backends.etcd.EtcdBackend',
        client=client,
    )

    def tearDown(self):
        super(EtcdBackendTest, self).tearDown()
        try:
            self.client.delete(self.path)
        except etcd.EtcdKeyNotFound:
            pass

    def create_config(self, value):
        value = self.serialize_config(value)
        self.client.set(self.path, value, ttl=EPHEMERAL_TEST_DATA_TTL)

    def update_config(self, new_value):
        new_value = self.serialize_config(new_value)
        self.client.set(self.path, new_value, ttl=EPHEMERAL_TEST_DATA_TTL)

    def delete_config(self):
        self.client.delete(self.path)



from . import config, utils


class Proxy:
    """Proxy class for different backends."""

    def __init__(self, backend):
        self._backend = backend

    @property
    def backend(self):
        """Readonly property for backend."""
        return self._backend

    @classmethod
    def configure(cls, backend_qualname, **backend_options):
        """Configure backend.

        :param backend_qualname: Backend class qualified name (dotted-name)
            e.g. ``distconfig.backends.zookeeper.ZooKeeperBackend``.
        :param logger: ``logging.Logger`` instance to use for logging.
        :param parser: Callable that accept the raw config data saved in
            the backend and should return the config data parsed,
            default: ``json.loads``.
        :param backend_options: Keyword arguments to pass to backend class.
        """
        backend_cls = utils.resolve_dotted_name(backend_qualname)
        backend = backend_cls(**backend_options)
        return cls(backend)

    def get_config(self, path, config_cls=config.Config):
        """Get configuration from path.

        :param path: Location of the configuration in the backend.
        :param config_cls: configuration class to return, default: :class:`distconfig.config.Config`.
        :return: ``config_cls`` instance.

        """
        data = self._backend.get(path)
        config = config_cls(data)
        self._backend.add_listener(config._invalidate)
        return config



import json
import unittest

import mock

from distconfig.backends import base


class FakeBackend(base.BaseBackend):

    def __init__(self, data, **kwargs):
        super(FakeBackend, self).__init__(**kwargs)
        self._data = data

    def get_raw(self, path):
        return self._data.get(path)


class BackendTestCase(unittest.TestCase):

    def setUp(self):
        self.value = {'foo': 'bar'}
        self.raw_value = json.dumps(self.value)
        self.backend = FakeBackend({
            '/some/path': self.raw_value
        })

    def test_backend_get(self):
        self.assertEqual(self.backend.get('/some/path'), self.value)

    def test_backend_get_unexistant_path(self):
        self.assertEqual(self.backend.get('/no/existing/path'), {})

    def test_add_listeners(self):
        callback = mock.Mock(return_value=None)
        self.backend.add_listener(callback)

        self.backend._notify_listeners(self.raw_value)

        callback.assert_called_once_with(self.value)

    def test_add_multiple_listeners(self):
        callback1 = mock.Mock(return_value=None)
        self.backend.add_listener(callback1)

        callback2 = mock.Mock(return_value=None)
        self.backend.add_listener(callback2)

        self.backend._notify_listeners(self.raw_value)

        callback1.assert_called_once_with(self.value)
        callback2.assert_called_once_with(self.value)

    def test_add_listeners_multiple_time(self):
        callback = mock.Mock(return_value=None)
        self.backend.add_listener(callback)
        self.backend.add_listener(callback)
        self.backend.add_listener(callback)

        self.backend._notify_listeners(self.raw_value)

        self.assertEqual(callback.call_count, 3)
        callback.assert_called_with(self.value)

    def test_remove_listeners(self):
        callback = mock.Mock(return_value=None)
        self.backend.add_listener(callback)

        self.backend.remove_listener(callback)

        self.backend._notify_listeners(self.raw_value)

        self.assertEqual(callback.call_count, 0)

    def test_remove_listeners_multiple_time(self):
        callback = mock.Mock(return_value=None)
        self.backend.add_listener(callback)
        self.backend.add_listener(callback)
        self.backend.add_listener(callback)

        self.backend.remove_listener(callback)
        self.backend.remove_listener(callback)
        self.backend.remove_listener(callback)

        self.backend._notify_listeners(self.raw_value)

        self.assertEqual(callback.call_count, 0)

    def test_remove_listeners_multiple_time_keeping_others(self):
        callback = mock.Mock(return_value=None)
        self.backend.add_listener(callback)
        self.backend.add_listener(callback)
        self.backend.add_listener(callback)

        self.backend.remove_listener(callback)

        self.backend._notify_listeners(self.raw_value)

        self.assertEqual(callback.call_count, 2)

    def test_remove_listeners_more_than_you_should(self):
        callback = mock.Mock(return_value=None)
        self.backend.add_listener(callback)

        self.backend.remove_listener(callback)

        with self.assertRaises(ValueError):
            self.backend.remove_listener(callback)

    def test_listener_exception(self):
        callback = mock.Mock(side_effect=Exception('foo'))
        self.backend.add_listener(callback)

        with self.assertRaises(Exception):
            self.backend._notify_listeners(self.raw_value)

    def test_listener_log_exception_on_error(self):
        logger_mock = mock.Mock()

        backend = FakeBackend(
            {'/some/path': self.raw_value},
            logger=logger_mock)

        callback = mock.Mock(side_effect=Exception('foo'))
        backend.add_listener(callback)

        with self.assertRaises(Exception):
            backend._notify_listeners(self.raw_value)

        self.assertTrue(logger_mock.exception.called)
